
from cython cimport boundscheck, wraparound, cdivision
from cython.view cimport contiguous
from cython.parallel cimport prange

from cal_tools.enums import BadPixels

ctypedef unsigned short raw_t
ctypedef float data_t
ctypedef unsigned char gain_t
ctypedef unsigned short cell_t
ctypedef unsigned int mask_t

cdef mask_t WRONG_GAIN_VALUE = BadPixels.WRONG_GAIN_VALUE, \
    VALUE_OUT_OF_RANGE = BadPixels.VALUE_OUT_OF_RANGE, \
    VALUE_IS_NAN = BadPixels.VALUE_IS_NAN


@boundscheck(False)
@wraparound(False)
def correct_lpd_frames(
    # (frame, x, y)
    raw_t[:, :, ::contiguous] in_raw,
    cell_t[::contiguous] in_cell,

    # (frame, x, y)
    data_t[:, :, ::contiguous] out_data,
    gain_t[:, :, ::contiguous] out_gain,
    mask_t[:, :, ::contiguous] out_mask,

    # (cell, x, y, gain)
    data_t[:, :, :, ::contiguous] ccv_offset,
    data_t[:, :, :, ::contiguous] ccv_noise,
    data_t[:, :, :, ::contiguous] ccv_gain,
    mask_t[:, :, :, ::contiguous] ccv_mask,

    # (frame, gain)
    int[:, ::contiguous] parallel_gain_indices,

    data_t threshold_sigma_high,
    data_t threshold_sigma_mid,

    int num_threads=1,
):
    cdef int frame, ss, fs
    cdef cell_t cell
    cdef data_t data
    cdef gain_t gain
    cdef mask_t mask

    cdef bint adaptive_gain = parallel_gain_indices is None

    for frame in prange(out_data.shape[0], nogil=True, num_threads=num_threads):
        cell = in_cell[frame]

        for ss in range(in_raw.shape[1]):
            for fs in range(in_raw.shape[2]):
                if adaptive_gain:
                    # Decode intensity and gain from raw data.
                    data = <data_t>(in_raw[frame, ss, fs] & 0xFFF)
                    gain = <gain_t>((in_raw[frame, ss, fs] & 0x3000) >> 12)

                else:
                    _parallel_gain_thresholding(
                        in_raw, parallel_gain_indices, ccv_noise,
                        threshold_sigma_high, threshold_sigma_mid,
                        frame, cell, ss, fs,
                        &data, &gain)

                if gain <= 2:
                    data = data - ccv_offset[cell, ss, fs, gain]
                    data = data * ccv_gain[cell, ss, fs, gain]
                    mask = ccv_mask[cell, ss, fs, gain]
                else:
                    data = 0.0
                    gain = 0
                    mask = WRONG_GAIN_VALUE

                if data > 1e7 or data < -1e7:
                    data = 0.0
                    mask = mask | VALUE_OUT_OF_RANGE

                out_data[frame, ss, fs] = data
                out_gain[frame, ss, fs] = gain
                out_mask[frame, ss, fs] = mask


@boundscheck(False)
@wraparound(False)
cdef inline void _parallel_gain_thresholding(
    raw_t[:, :, ::contiguous] in_raw,
    int[:, ::contiguous] parallel_gain_indices,
    data_t[:, :, :, ::contiguous] ccv_noise,
    data_t sigma_high, data_t sigma_mid,
    int frame, int cell, int ss, int fs,
    data_t* data_ptr, gain_t* gain_ptr
) noexcept nogil:
    cdef int frame_high, frame_mid, frame_low
    cdef data_t data_high, data_mid, data_low
    cdef data_t threshold_high, threshold_mid

    # Obtain indices to this pixel in each of three gain images.
    frame_high = parallel_gain_indices[frame, 0]
    frame_mid = parallel_gain_indices[frame, 1]
    frame_low = parallel_gain_indices[frame, 2]

    if (
        ((in_raw[frame_high, ss, fs] & 0x3000) >> 12) == 0 and
        ((in_raw[frame_mid, ss, fs] & 0x3000) >> 12) == 1 and
        ((in_raw[frame_low, ss, fs] & 0x3000) >> 12) == 2
    ):
        # Verify that this pixel is recorded in the correct gain stage
        # in each of the gain images. Memory cells in the transition
        # regions between the gains can sometimes end up in the wrong
        # one.

        # Decode intensity in every gain stage.
        data_high = <data_t>(in_raw[frame_high, ss, fs] & 0xFFF)
        data_mid = <data_t>(in_raw[frame_mid, ss, fs] & 0xFFF)
        data_low = <data_t>(in_raw[frame_low, ss, fs] & 0xFFF)

        # Compute thresholds based on noise level.
        threshold_high = 4096 - sigma_high * ccv_noise[cell, ss, fs, 0]
        threshold_mid = 4096 - sigma_mid * ccv_noise[cell, ss, fs, 1]

        # Pick the optimal gain stage for this pixel.
        if data_mid > threshold_mid:
            data_ptr[0] = data_low
            gain_ptr[0] = 2
        elif data_high > threshold_high:
            data_ptr[0] = data_mid
            gain_ptr[0] = 1
        else:
            data_ptr[0] = data_high
            gain_ptr[0] = 0

    else:
        # Using an invalid gain stage triggers bad pixel masking later
        # in the correction kernel.
        gain_ptr[0] = 4
