import argparse
import os
import shutil
import sys
from datetime import datetime
from pathlib import Path

import nbformat
from env_cache import EnvsManager, PyenvEnvMaker
from nbparameterise import extract_parameters, parameter_values, replace_definitions

from cal_tools.tools import CalibrationMetadata
from .calibrate import (
    JobChain, SlurmOptions, run_finalize, get_pycalib_version,
)

# This function is copied and modified from Python 3.8.10
# Copyright © 2001-2022 Python Software Foundation; All Rights Reserved
# Used under the PSF license - https://docs.python.org/3/license.html
def copytree_no_metadata(src, dst, ignore):
    """Copy directory contents, without permissions or extended attributes

    Copying the GPFS ACLs can cause problems when creating new files in
    the directory, so we avoid that. shutil.copytree can skip copying file
    metadata, but not directory metadata:
    https://github.com/python/cpython/issues/89721

    This is also simplified by removing options we're not using.
    """
    with os.scandir(src) as itr:
        entries = list(itr)
    if ignore is not None:
        ignored_names = ignore(os.fspath(src), [x.name for x in entries])
    else:
        ignored_names = set()

    os.makedirs(dst)
    errors = []

    for srcentry in entries:
        if srcentry.name in ignored_names:
            continue
        srcname = os.path.join(src, srcentry.name)
        dstname = os.path.join(dst, srcentry.name)
        try:
            if srcentry.is_dir():
                copytree_no_metadata(srcentry, dstname, ignore)
            else:
                # Will raise a SpecialFileError for unsupported file types
                shutil.copyfile(srcentry, dstname)
        # catch the Error from the recursive copytree so that we can
        # continue with other files
        except shutil.Error as err:
            errors.extend(err.args[0])
        except OSError as why:
            errors.append((srcname, dstname, str(why)))

    if errors:
        raise shutil.Error(errors)
    return dst

def update_notebooks_params(working_dir: Path, new_values):
    """Modify parameters in copied notebooks"""
    for nb_file in working_dir.glob('*.ipynb'):
        nb = nbformat.read(nb_file, as_version=4)
        params = extract_parameters(nb, lang='python')
        params = parameter_values(params, **new_values)
        new_nb = replace_definitions(nb, params, execute=False, lang='python')
        nbformat.write(new_nb, nb_file)


def get_python(args, py_version):
    if args.env_cache:
        reqs = (args.from_dir / 'requirements.txt').read_text('utf-8')
        reqs = munge_requirements(reqs)
        env_mgr = EnvsManager(
            Path(args.env_cache), PyenvEnvMaker()
        )
        return env_mgr.get_env(py_version, reqs).resolve() / 'bin' / 'python'
    elif args.python:
        return args.python.resolve()
    return Path(sys.executable)


def munge_requirements(reqs: str):
    # Ugly workaround - these internal packages are dependencies of
    # pycalibration, but 'pip freeze' shows them differently to how they are
    # specified in setup.py, which causes conflicts in 'pip install'. For now,
    # so we can recalibrate old data, we'll cut them out of the requirements
    # list and rely on pycalibration specifying them as pinned dependencies.
    reqs = reqs.splitlines(keepends=False)
    reqs = [r for r in reqs if not r.startswith(
        ('iCalibrationDB', 'XFELDetectorAnalysis')
    )]
    return '\n'.join(reqs)


def script_fix(cal_work_dir: Path):
    # Workaround: until May 2024, this script contained a leftover & unnecessary
    # 'module load anaconda/3' command. This module is not available on EL9.
    # We don't use the module at all, so we can safely cut the command out.
    script = cal_work_dir / "pycalib-run-nb.sh"
    if not script.is_file():
        return

    lines = script.read_text().splitlines(keepends=True)
    script.write_text(''.join(
        l for l in lines if l.strip().split() != ['module', 'load', 'anaconda/3']
    ))


def main(argv=None):
    ap = argparse.ArgumentParser()
    ap.add_argument("from_dir", type=Path, help="A directory containing steps.json")

    env_args = ap.add_mutually_exclusive_group()
    env_args.add_argument("--python", type=Path,
                          help="Path to Python executable to run notebooks")
    env_args.add_argument(
        "--env-cache",
        help="Make/reuse a virtualenv in this cache directory with the Python "
             "version & library versions these notebooks previously used. "
             "This requires the pyenv command to be available."
    )
    ap.add_argument("--out-folder", help="Directory to put output data")
    ap.add_argument("--report-to", help="Location to save PDF report")
    ap.add_argument("--slurm-partition", help="Submit jobs in this Slurm partition")
    ap.add_argument("--slurm-mem", type=int, help="Requested node RAM in GB")
    ap.add_argument('--no-cluster-job', action="store_true",
                    help="Run notebooks here, not in cluster jobs")
    args = ap.parse_args(argv)

    if not (args.from_dir / 'steps.json').is_file():
        sys.exit(f"Directory {args.from_dir} does not contain steps.json")

    start_time = datetime.now()
    run_uuid = f"t{start_time:%y%m%d_%H%M%S}"

    parameters = CalibrationMetadata(args.from_dir)['calibration-configurations']
    karabo_id = parameters['karabo-id']

    out_folder = parameters['out-folder']
    params_to_set = {'metadata_folder': "."}
    if args.out_folder:
        out_folder = parameters['out-folder'] = os.path.abspath(args.out_folder)
        params_to_set['out_folder'] = out_folder

    if args.report_to:
        report_to = Path(args.report_to).with_suffix('.pdf').absolute()
    else:  # Default to saving report in output folder
        report_to = Path(out_folder, f'{karabo_id}-repeat-{run_uuid}.pdf')

    # Copy working directory to new location
    cal_work_dir = report_to.with_suffix('')
    copytree_no_metadata(
        args.from_dir, cal_work_dir, ignore=shutil.ignore_patterns('slurm-*.out')
    )
    print(f"New working directory: {cal_work_dir}")

    script_fix(cal_work_dir)  # Strip out 'module load anaconda/3'

    # Update metadata YAML file & notebooks with any changes
    cal_metadata = CalibrationMetadata(cal_work_dir)
    cal_metadata['calibration-configurations'] = parameters
    cal_metadata['report-path'] = str(report_to)
    cal_metadata.save()

    update_notebooks_params(cal_work_dir, params_to_set)

    # finalize & some notebooks expect yaml metadata in the output folder
    Path(out_folder).mkdir(parents=True, exist_ok=True)
    shutil.copy(cal_work_dir / 'calibration_metadata.yml', out_folder)

    py_version = cal_metadata.get('python-environment', {}).get('python-version')
    job_chain = JobChain.from_dir(
        cal_work_dir, python=get_python(args, py_version)
    )
    if args.no_cluster_job:
        job_chain.run_direct()
        joblist = []
    else:
        # The queries to look up constants should all be the same as those
        # from the previous calibration - tell CalParrot to warn if not.
        env = os.environ.copy()
        env['CALPARROT_NEW_QUERY'] = 'warn'

        joblist = job_chain.submit_jobs(SlurmOptions(
            partition=args.slurm_partition,
            mem=args.slurm_mem,
        ), env=env)

    fmt_args = {'cal_work_dir': cal_work_dir,
                'out_path': out_folder,
                'version': get_pycalib_version(),
                'report_to': report_to,
                'in_folder': parameters['in-folder'],
                'request_time': start_time.strftime('%Y-%m-%dT%H:%M:%S'),
                'submission_time': start_time.strftime('%Y-%m-%dT%H:%M:%S'),
                }

    joblist.append(run_finalize(
        fmt_args=fmt_args,
        cal_work_dir=cal_work_dir,
        job_list=joblist,
        sequential=args.no_cluster_job,
        partition=args.slurm_partition
    ))

    if any(j is not None for j in joblist):
        print("Submitted the following SLURM jobs: {}".format(",".join(joblist)))


if __name__ == '__main__':
    sys.exit(main())
